import hashlib
from pathlib import Path
import os
from time import perf_counter
from tqdm import tqdm
from random import choice, choices, randint
from string import ascii_letters
from send2trash import send2trash
import json
from datetime import datetime as dt, timezone
import click
from duple.decorators import log_func_time, log_func_with_args
from duple.info import LOGS_PATH

from typing import Iterator
from time import sleep


def get_latest_file(inpath: str, filter: str = None, recurse: bool = False) -> str:
    files = list()

    path: Path
    path = inpath
    if isinstance(path, str):
        path = Path(path)

    def add_file(file: Path):
        if file.is_file() and (filter is None or filter in file.name):
            files.append(file)

    if recurse:
        for r, fd, fs in os.walk(path):
            for f in fs:
                file: Path = Path(r).joinpath(f)
                add_file(file)
    else:
        for f in os.listdir(path):
            file = Path(path).joinpath(f)
            add_file(file)

    if files is None:
        raise Exception(FileNotFoundError)

    atimes = {str(path.absolute()): path.stat().st_atime for path in files}
    min_file = max(atimes.items(), key=lambda x: x[1])

    # logger.info(f'latest accessed file: {min_file[0]}')
    return min_file[0]


def follow_log(level: str = "DEBUG", filter: str = None):
    level = level.upper()
    levels = {
        "NOTSET": {"color": "white", "value": 0},
        "DEBUG": {"color": "green", "value": 10},
        "INFO": {"color": "magenta", "value": 20},
        "WARNING": {"color": "yellow", "value": 30},
        "ERROR": {"color": "bright_red", "value": 40},
        "CRITICAL": {"color": "bright_red", "value": 50},
    }

    file_path = get_latest_file(LOGS_PATH, filter=".jsonl")
    # logger.debug(f'Latest log file is {file_path}')
    try:
        with open(file_path, "r") as file:
            line = file.readline()
            print(line)

            for line in follow(file):
                parsed = json.loads(line)
                style_ts = click.style(f"[{parsed['timestamp']}]", fg="cyan")
                style_level = click.style(f"[{parsed['level']}]", fg=levels[parsed["level"]]["color"])
                style_module = click.style(parsed["module"], fg="white")
                style_function = click.style(parsed["function"], fg="white")
                style_message = click.style(parsed["message"], fg=levels[parsed["level"]]["color"])
                style_msg_pointer = click.style("->", fg="bright_blue")

                if levels[parsed["level"]]["value"] >= levels[level]["value"] and (filter is None or filter in line):
                    click.secho(
                        f"{style_ts} {style_level} {style_module}.{style_function} {style_msg_pointer} {style_message}"
                    )

                # if Path(file_path).is_file():
                #     new_file_path = get_latest_file(LOGS_PATH, filter = '.jsonl')
                #     if file_path != new_file_path:
                #         #logger.debug(f'Log file rollover detected, switching to newest log: {file_path}')
                #         file_path = new_file_path
    except:  # noqa E722
        pass


def follow(file, sleep_sec=0.1) -> Iterator[str]:
    """Yield each line from a file as they are written.
    `sleep_sec` is the time to sleep after empty reads."""
    line: str = str()
    while True:
        tmp = file.readline()
        if tmp is not None and tmp != "":
            line += tmp
            if line.endswith("\n"):
                yield line
                line = ""
        elif sleep_sec:
            sleep(sleep_sec)


def dump_dict_to_json(input: dict, filename: str, indent: int = 4):
    with open(filename, "w") as f:
        json.dump(input, f, indent=indent)


def get_available_hashes() -> list:
    hashes = list(hashlib.algorithms_available)
    hashes.remove("shake_128")
    hashes.remove("shake_256")
    hashes.sort()
    return hashes


@log_func_with_args
def gen_test_files(test_path_root: Path, numdirs: int, numfiles: int, max_file_size: int):
    if not Path(test_path_root).exists():
        os.mkdir(test_path_root)

    file_sizes = [x for x in range(max_file_size)]
    # data = [os.urandom(file_size) for _ in range(int(numfiles/2))]
    data = [os.urandom(choice(file_sizes)) for _ in range(int((numfiles * numdirs) / 2))]

    for i in tqdm(range(numdirs), desc="making directories", position=0):
        dir_path = test_path_root.joinpath("".join(choices(ascii_letters, k=randint(3, 30))))
        dir_path2 = dir_path.joinpath("".join(choices(ascii_letters, k=randint(3, 30))))

        if not Path(dir_path).exists():
            os.mkdir(dir_path)

        if not Path(dir_path2).exists():
            os.mkdir(dir_path2)

        for j in range(numfiles):
            file_name = "".join(choices(ascii_letters, k=randint(3, 30)))
            file_path = Path(choice([dir_path, dir_path2])).joinpath(file_name + ".txt")

            if file_path.exists():
                continue

            with open(file_path, "wb") as f:
                f.write(choice(data))


def get_hash(path: str, hash: str) -> str | str:
    try:
        with open(path, "rb") as f:
            digest = hashlib.file_digest(f, hash)
        return digest.hexdigest(), path
    except PermissionError:
        return "", path


@log_func_with_args
def timed_get_hash(hash: str, path: str) -> float:
    start = perf_counter()
    get_hash(path, hash)
    finish = perf_counter()
    return hash, finish - start


def get_max_workers():
    return int(os.cpu_count() * 0.75) + 1


def tree(dir_path: Path, prefix: str = ""):
    """A recursive generator, given a directory Path object
    will yield a visual tree structure line by line
    with each line prefixed by the same characters
    """

    # prefix components:
    space = "    "
    branch = "│   "
    # pointers:
    tee = "├── "
    last = "└── "

    contents = list(dir_path.iterdir())
    # contents each get pointers that are ├── with a final └── :
    pointers = [tee] * (len(contents) - 1) + [last]
    for pointer, path in zip(pointers, contents):
        yield [prefix, pointer, path.name, path.is_dir()]
        if path.is_dir():  # extend the prefix and recurse:
            extension = branch if pointer == tee else space
            # i.e. space because last, └── , above so no more |
            yield from tree(path, prefix=prefix + extension)


def generate_options_list(flags: list) -> list[tuple]:
    options = list()
    if flags[0]:
        option = ("depth", True)
        options.append(option)
    if flags[1]:
        option = ("depth", False)
        options.append(option)
    if flags[2]:
        option = ("namelength", True)
        options.append(option)
    if flags[3]:
        option = ("namelength", False)
        options.append(option)
    if flags[4]:
        option = ("atime", True)
        options.append(option)
    if flags[5]:
        option = ("atime", False)
        options.append(option)
    if flags[6]:
        option = ("ctime", True)
        options.append(option)
    if flags[7]:
        option = ("ctime", False)
        options.append(option)
    if flags[8]:
        option = ("mtime", True)
        options.append(option)
    if flags[9]:
        option = ("mtime", False)
        options.append(option)

    return options


@log_func_with_args
def remove_empty_directories(path: str, verbose: bool = False) -> None:
    for root, folders, files in tqdm(
        os.walk(path), desc="Removing .DS_Store files from empty directories", leave=False
    ):
        if ".DS_Store" in files and len(files) == 1:
            temp_path = str(Path(root).joinpath(".DS_Store").absolute())
            send2trash(temp_path)

    dirs = list()
    for root, folders, files in os.walk(os.getcwd()):
        if not files and not folders:
            dirs.append(root)

    dirs = sorted(dirs, key=lambda x: -1 * len(Path(x).parents))

    if len(dirs) == 0:
        return

    # [ 100.0%] deleted             90 Bytes  nZp/enRmIFgQjR/iZSR.txt

    if verbose:
        cnt = 0
        for dir in dirs:
            cnt += 1
            completestyle = click.style(f"[{cnt / len(dirs) * 100: 6.1f}%]", fg="cyan")
            deldirstyle = click.style("deleting empty directory: ", fg="yellow")
            click.secho(f"{completestyle} {deldirstyle} {dir}")
            send2trash(dir)
    else:
        for dir in tqdm(dirs, desc="Removing empty directories"):
            send2trash(dir)


@log_func_time
def create_output(inputs: list, output_dir: str, dupes: list, ignored_files: list, summary_statistics: dict) -> None:
    max_len_key = max([len(k) for k in summary_statistics.keys()])
    max_len_val = max([len(str(v)) for v in summary_statistics.values()])

    section_divider = f"\n{''.rjust(max_len_key + max_len_val + 10,'-')}\n"

    # output_json = Path(output_dir).joinpath('duple.json')
    output_delete = Path(output_dir).joinpath("duple.delete")

    for key, value in summary_statistics.items():
        if "(seconds)" in key:
            summary_statistics[key] = f"{round(value, 5):.5f}"
            value = summary_statistics[key]

        click.secho(
            f'{key.ljust(max_len_key + max_len_val - len(str(value)) + 10, ".")}{click.style(str(value), fg = "green")}'
        )

    click.secho()
    click.secho(
        "Open the `output summary results` file listed above with a text editor for review",
        fg="yellow",
    )
    click.secho("Once review and changes are complete, run `duple rm`", fg="yellow")

    # dump_dict_to_json(all_files, output_json.absolute())

    with open(output_delete, "w", encoding="utf-8") as f:
        f.write(
            f"Duple Report Generated on {dt.now(timezone.utc).astimezone().isoformat()}, commanded by user: {os.getlogin()}"
        )
        f.write(section_divider)
        f.write("Summary Statistics:\n")
        for key, value in summary_statistics.items():
            f.write(f'{key.ljust(max_len_key + max_len_val - len(str(value)) + 10,".")}{str(value)}\n')
        f.write(section_divider)
        f.write("Inputs (True = minimum, False = Maximum): \n")
        for input in inputs:
            f.write(f"{input[0]} = {input[1]}\n")

        f.write(section_divider)
        f.write("Outputs:\n")
        # f.write(f"{str(output_json.absolute())}\n")
        f.write(f"{str(output_delete.absolute())}\n")

        f.write(section_divider)
        f.write("Instructions to User:\n")
        f.write(
            "The sections below describe what action duple will take when 'duple rm' is commanded."
            " The first column is the flag that tells duple what to do:\n"
            "\tORIGINAL   : means duple will take no action for this file, listed only as a reference to the user\n"
            "\tDUPLICATE  : means duple will send this file to the trash can or recycling bin, if able\n"
        )

        f.write(section_divider)
        f.write("Duplicate Results:\n")
        if len(dupes) == 0:
            f.write("No duplicates found.\n")

        for line in dupes:
            f.write(f"{line}\n")

        f.write(section_divider)
        f.write("Ignored Files in Scan:\n")
        for line in ignored_files:
            f.write(f"{line}\n")


@log_func_with_args
def get_delete_paths(path: Path) -> dict:
    if not duple_outputs_exist(path):
        return

    with open(path.joinpath("duple.delete"), "r") as f:
        items = f.read().splitlines()

    results = dict()
    flag = False
    for item in items:
        if item == "Duplicate Results:":
            flag = True

        if flag and "----" in item:
            break

        if flag and item.split("|")[0].strip() in ["DUPLICATE", "ORIGINAL"]:
            item_split = item.split("|")
            action = item_split[0].strip()
            size = item_split[1]  # .strip()
            path = item_split[2][1:].strip()
            results[path] = {"type": action, "size": size, "path": path}

    return results


def duple_outputs_exist(path: Path) -> bool:
    if not path.joinpath("duple.delete").exists():
        click.secho("duple.delete do no exists - run duple scan to create this file")
        return False
    return True


def delete_logs() -> str:
    files = os.listdir(LOGS_PATH)
    for file in files:
        p = Path(LOGS_PATH).joinpath(file)
        p.unlink()

    p = Path(LOGS_PATH).joinpath("log.jsonl")
    with open(p, "a"):
        pass

    p = p.parent.joinpath("log.log")
    with open(p, "a"):
        pass

    return f"Deleted Log Files: {files}"



import cProfile
import functools
import pstats
from time import perf_counter
from uuid import uuid1
from duple.app_logging import logger, setup_logging

setup_logging()


def profile(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        profiler = cProfile.Profile()
        profiler.enable()
        try:
            retval = func(*args, **kwargs)
        finally:
            profiler.disable()
            with open("profile.out", "w") as profile_file:
                stats = pstats.Stats(profiler, stream=profile_file)
                stats.print_stats()
        return retval

    return inner


def log_func_with_args(func):
    start = perf_counter()

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        funcname = f"{func.__module__}/{func.__qualname__}"
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        logger.debug(f"{funcname} called with args {signature}")
        try:
            logger.debug(f"{funcname} START", extra={"funcname": funcname})
            result = func(*args, **kwargs)
            logger.debug(f"{funcname} FINISH ({(perf_counter() - start): .5f})", extra={"funcname": funcname})
            return result
        except Exception as e:
            logger.exception(f"Exception raised in function {funcname}, exception: {str(e)}")
            raise e

    return wrapper


def log_func_time(func):
    start = perf_counter()

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        funcname = f"{func.__module__}/{func.__qualname__}"
        try:
            logger.debug(f"{funcname} START")
            result = func(*args, **kwargs)
            logger.debug(
                f"{funcname} FINISH elapsed time: {(perf_counter() - start):.5f}", extra={"funcname": funcname}
            )
            return result
        except Exception as e:
            logger.exception(
                f"Exception raised in function {funcname}, exception: {str(e)}", extra={"funcname": funcname}
            )
            raise e

    return wrapper


def log_func_start_finish_flags(func):
    start_count = perf_counter()

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        funcname = f"{func.__module__}/{func.__qualname__}"
        flag = "".ljust(20, "=")
        start = f"START {funcname}".center(40)
        finish = f"FINISH {funcname}".center(40)
        uid = uuid1().hex
        try:
            logger.info(f"{flag} {start} {flag} {uid=}", extra={"funcname": funcname})
            result = func(*args, **kwargs)
            logger.info(
                f"{flag} {finish} {flag} {uid=} elapsed time: {(perf_counter() - start_count):.5f} seconds",
                extra={"funcname": funcname},
            )

            return result
        except Exception as e:
            logger.exception(f"Exception raised in function {funcname}, exception: {str(e)}")
            raise e

    return wrapper



from enum import Enum, auto


class Status(Enum):
    IGNORED = auto()
    POTENTIAL_DUPLICATE = auto()
    DUPLICATE = auto()
    ORIGINAL = auto()
    NOT_ANALYZED = auto()

    def longest_status():
        statuses = [value.name for value in Status]
        statuses.remove("POTENTIAL_DUPLICATE")
        statuses.remove("NOT_ANALYZED")

        result = max([len(str(value)) for value in statuses])
        return result



from dataclasses import dataclass, field
from pathlib import Path
from duple.file import File
from duple.status import Status
from duple.dispocode import DispoCode
from tqdm.contrib.concurrent import process_map
from itertools import repeat
from duple.library import get_hash
from humanize import naturalsize
import os
from collections import Counter
from duple.decorators import log_func_time
from duple.app_logging import logger


"""
files is the collector the file class, files will:
-keep the all file objects up to date with the analysis results
-provide output file for duplicate review by user
-provide output of all files analyzed
-generate file objects from a list of paths (strings or Paths) and add to the collector
-identify duplicate/original files
    -preprocess on file size to reduce the number of hashes required
    -hash remainging files
    -identify original file based on input options
-ingest the output reviewed and updated by the user
-provide an iterable that cycles through all of the files and gives an action to take
"""


@dataclass
class Files(dict):
    duplicates: dict = field(init=False, default_factory=dict)
    test_functions: list = field(init=False, default_factory=list)
    max_workers: int = field(default=int(os.cpu_count() * 0.75) + 1)
    chunksize: int = field(default=2)
    hashalgo: str = field(default="sha256")
    duplicate_file_size: int = field(init=False, default=0)
    total_file_size: int = field(init=False, default=0)
    duplicate_file_count: int = field(init=False, default=0)
    duplicate_group_count: int = field(init=False, default=0)
    file_count: int = field(init=False, default=0)
    ignored_files: int = field(init=False, default=0)

    def _test_fun(self, attribute: str, minimum: bool, files: list[File]) -> list[File]:
        """
        test_fun returns a list of file objects that pass the test defined by attribute and minimum

        Args:
            attribute (str): attribute to test
            minimum (bool): true if attribute should be the minimum, false if the attribute should be the maxium
            files (list[File]): list of file objects to check

        Returns:
            list[File]: list of file objects that pass the test
        """

        temp = dict()
        for file in files:
            # print(file.__dict__)
            temp[str(file.path)] = file.__dict__[attribute]

        target = max(temp.items(), key=lambda x: x[1])
        if minimum:
            target = min(temp.items(), key=lambda x: x[1])

        target = temp[target[0]]

        return [file for file in files if file.__dict__[attribute] == target]

    @log_func_time
    def read_paths(self, paths: list):
        for path in paths:
            f = File(path)
            self[f.path] = f

    def get_paths(self) -> list:
        return list(self.keys())

    def get_status(self) -> dict:
        return {key: value.status for key, value in self.items()}

    @log_func_time
    def pre_process_files(self) -> None:
        """
        Eliminate files with unique sizes and ignored files, these can not be duplicates
        """
        sizes = dict()
        tempfile: File
        for tempfile in self.values():
            if tempfile.status == Status.IGNORED:
                continue

            if tempfile.size not in sizes.keys():
                sizes[tempfile.size] = list()

            sizes[tempfile.size].append(tempfile)

        itemlist: list
        for itemlist in sizes.values():
            item: File
            for item in itemlist:
                if len(itemlist) > 1:
                    self[item.path].status = Status.POTENTIAL_DUPLICATE
                else:
                    self[item.path].status = Status.IGNORED
                    self[item.path].dispocode = DispoCode.UNIQUE_FILE_SIZE

    @log_func_time
    def process_files(self) -> None:
        logger.debug("ID of Potential Duplicates START")
        potential_duplicates = list()
        tempfile: File
        for tempfile in self.values():
            if tempfile.status == Status.POTENTIAL_DUPLICATE:
                potential_duplicates.append(str(tempfile.path))
        logger.debug("ID of Potential Duplicates FINISH")

        logger.debug("Hashing Files START")
        hashes = process_map(
            get_hash,
            potential_duplicates,
            repeat(self.hashalgo),
            max_workers=self.max_workers,
            chunksize=self.chunksize,
            desc="hashing files",
        )
        logger.debug("Hashing Files FINISH")

        logger.debug("Recording Hash Values to Files START")
        for hash, path in hashes:
            if not hash:
                self[Path(path)].status = Status.IGNORED
                self[Path(path)].dispocode = DispoCode.PERMISSION_DENIED
            else:
                self[Path(path)].hash = hash
        logger.debug("Recording Hash Values to Files FINISH")

        logger.debug("Making Duplicates List START")
        for value in self.values():
            if value.hash not in self.duplicates.keys() and value.status == Status.POTENTIAL_DUPLICATE:
                self.duplicates[value.hash] = list()
            if value.status == Status.POTENTIAL_DUPLICATE:
                self.duplicates[value.hash].append(value)
        logger.debug("Making Duplicates List FINISH")

        logger.debug("Marking Ignored/Duplicate START")
        for items in self.duplicates.values():
            for item in items:
                self[item.path].twins = len(items) - 1  # [tfile.path for tfile in items if tfile != item]
                if len(items) == 1:
                    self[item.path].status = Status.IGNORED
                    self[item.path].dispocode = DispoCode.UNIQUE_HASH
                if len(items) > 1:
                    self[item.path].status = Status.DUPLICATE
        logger.debug("Marking Ignored/Duplicate FINISH")

        logger.debug("Finalizing Duplicates START")
        self.duplicates = {key: value for key, value in self.duplicates.items() if len(value) > 1}
        logger.debug("Finalizing Duplicates FINISH")

    @log_func_time
    def determine_originals(self, options: list):
        """
        options = tuple(attribute: str, minimum: bool)
        """
        if not isinstance(options, list):
            raise Exception(TypeError, f"options must be type = list, provided type = {type(options)}")

        for option in options:
            if not isinstance(option, tuple):
                raise Exception(TypeError, f"options must be a list of tuples, provided a list of {type(option)}")

            if option[0] not in File.get_available_option_attributes():
                raise Exception(
                    ValueError, f"first item in tuple must be in the list {File.get_available_option_attributes()}"
                )

            if option[1] not in [True, False]:
                raise Exception(ValueError, "second item in tuple must be True or False")

        for twins in self.duplicates.values():
            result = twins  # deepcopy(twins)

            for attribute, minimum in options:
                result = self._test_fun(attribute, minimum, result)

            original = result[0]
            self[original.path].status = Status.ORIGINAL

            for item in twins:
                if item.path != original.path:
                    self[item.path].status = Status.DUPLICATE

        self.calculate_totals()

    @log_func_time
    def create_duplicate_output(self) -> list:
        lines = list()
        for twins in self.duplicates.values():
            for file in twins:
                line = f"{str(file.status.name).ljust(Status.longest_status() + 2)}|"
                line += f"{naturalsize(file.size).center(13)}|"
                line += f" {file.path}"
                lines.append(line)
            lines.append("")

        return lines

    @log_func_time
    def create_ignored_files_output(self) -> list:
        lines = list()
        file: File
        logger.debug("Called Create Ignored Files Output")
        for file in self.values():
            if file.status.name == Status.IGNORED.name:
                line = f"{str(file.status.name).ljust(Status.longest_status() + 2)}|"
                line += f"{naturalsize(file.size).center(13)}|"
                line += f"{str(file.dispocode.name).center(DispoCode.longest_code() + 3)}|"
                line += f" {file.path}"
                lines.append(line)
        return lines

    def perform_analysis(self) -> None:
        self.pre_process_files()
        self.process_files()

    @log_func_time
    def calculate_totals(self):
        c = Counter([file.status for file in self.values()])

        self.duplicate_group_count = len(self.duplicates.keys())
        self.file_count = len(self.keys())
        self.duplicate_file_count = c[Status.DUPLICATE]
        self.ignored_files = c[Status.IGNORED]

        for file in self.values():
            if file.status == Status.DUPLICATE:
                self.duplicate_file_size += file.size
            self.total_file_size += file.size



from enum import Enum, auto


class DispoCode(Enum):
    UNIQUE_FILE_SIZE = auto()
    UNIQUE_HASH = auto()
    PERMISSION_DENIED = auto()
    NOT_SET = auto()

    def longest_code():
        codes = [value.name for value in DispoCode]

        result = max([len(str(value)) for value in codes])
        return result



from time import perf_counter
import click
from tqdm.contrib.concurrent import process_map
from itertools import repeat
from duple.__version__ import __version__
from duple.library import (
    get_available_hashes,
    timed_get_hash,
    get_max_workers,
    gen_test_files,
    tree,
    generate_options_list,
    create_output,
    remove_empty_directories,
    duple_outputs_exist,
    get_delete_paths,
    get_latest_file,
    follow_log,
    delete_logs,
)
import os
from pathlib import Path
import sys
from collections import Counter
from tqdm import tqdm
from duple.files import Files
from send2trash import send2trash
from humanize import naturalsize
from duple.app_logging import setup_logging, logger
from duple.decorators import log_func_with_args, log_func_start_finish_flags
from duple.info import LOGS_PATH

"""
To Do:
1. Add filtering capability for files (only pictures or files after a certain date)
2. Add StdIn support for scan
"""
"""
Only click interface functions will be in this file
"""


@click.group()
def cli():
    setup_logging()
    pass


@cli.command()
def wherelog():
    """print the path to the logs"""
    click.secho(f"{get_latest_file(LOGS_PATH, filter = '.jsonl')}")


@cli.command()
@click.option(
    "--level",
    "-l",
    type=click.STRING,
    default="DEBUG",
    help="Will only show specified level of log messages (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
)
@click.option("--filter", "-f", type=click.STRING, default=None)
def followlog(level: str = "DEBUG", filter: str = None):
    """follow the log until user interupts (ctrl-c), (just like "tail -f")"""

    while True:
        try:
            follow_log(level, filter)
        except:  # noqa E722
            pass


@cli.command()
def reset_logs():
    """empty the log files"""
    delete_logs()


@cli.command()
@click.argument("path", type=click.STRING)
@log_func_start_finish_flags
def hash_stats(path: str):
    """
    report hashing times for each available hashing algorithm on the specified file

    Args:
        path (str): path to file to hash
    """
    hashes = get_available_hashes()
    max_len_hash = max([len(hash) for hash in hashes]) + 2

    hash_times = process_map(
        timed_get_hash, hashes, repeat(path), max_workers=get_max_workers(), chunksize=1, desc="Hashing Files..."
    )

    hash_times = {k: v for k, v in hash_times}
    hash_times = sorted(hash_times.items(), key=lambda x: x[1])

    click.echo("Order = fastest > slowest")
    for hash, elapsed_time in hash_times:
        click.echo(f"{hash.ljust(max_len_hash)} {elapsed_time :8.6f} sec")


@cli.command()
@click.option(
    "--test_path", "-tp", type=click.Path(), default=os.getcwd(), help="path where the test directories will be created"
)
@click.option(
    "--number_of_directories", "-nd", type=click.INT, default=3, help="number of directories to make for the test"
)
@click.option(
    "--number_of_files",
    "-nf",
    type=click.INT,
    default=3,
    help="number of files to make in each top level directory, spread across the directories",
)
@click.option("--max_file_size", "-fs", type=click.INT, default=1024, help="file size to create in bytes")
@click.option("--print_tree", "-pt", is_flag=True, help="print tree with results")
@log_func_start_finish_flags
def make_test_files(
    test_path: Path, number_of_directories: int, number_of_files: int, max_file_size: int, print_tree: bool = False
):
    """make test files to learn or test with duple"""
    test_path = Path(test_path)
    gen_test_files(test_path, number_of_directories, number_of_files, max_file_size)

    if print_tree:
        click.secho()
        for line in tree(test_path):
            prefix_pointer_style = click.style("".join(line[:2]), fg="white")
            if line[-1]:
                name_style = click.style(line[2], fg="cyan")
            else:
                name_style = click.style(line[2], fg="white")
            click.secho(prefix_pointer_style + name_style)
        click.secho()


@cli.command()
@click.option("--verbose", "-v", is_flag=True, help="be more verbose during execution")
@click.option("--dry_run", "-dr", is_flag=True, help="Perform dry run, do everything except deleting files")
@click.option("--leave_empty_dirs", "-led", is_flag=True, help="Do not delete empty directories/folders")
@log_func_start_finish_flags
def rm(verbose: bool, dry_run: bool, leave_empty_dirs: bool) -> None:
    """
    rm sends all 'duplicate' files specified in duple.delete to the trash folder
    """
    logger.debug('Starting "duple rm"')

    message_margin = 16
    path = Path(os.getcwd())

    if not duple_outputs_exist(path):
        logger.debug('duple.delete does not exist, exiting "duple rm"')
        return

    files = get_delete_paths(path)

    delete_style = ""
    keep_style = ""

    if verbose:
        delete_style = click.style("deleted".ljust(message_margin), fg="yellow")
        keep_style = click.style("kept".ljust(message_margin), fg="green")

    if dry_run:
        delete_style = click.style("will delete".ljust(message_margin), fg="yellow")
        keep_style = click.style("will keep".ljust(message_margin), fg="green")
        verbose = True

    if not verbose:
        for file, data in tqdm(files.items()):
            if data["type"] == "DUPLICATE" and Path(file).exists():
                send2trash(file)

    if verbose:
        for i, (file, data) in enumerate(files.items()):
            completion_style = click.style(f"[{(i / len(files.items()) * 100) : 6.1f}%]", fg="cyan")

            if not dry_run and Path(file).exists() and data["type"] == "DUPLICATE":
                send2trash(file)

            if dry_run or verbose:
                if data["type"] == "ORIGINAL":
                    click.secho(f'{completion_style} {keep_style} {data["size"]}{file}')
                if data["type"] == "DUPLICATE":
                    click.secho(f'{completion_style} {delete_style} {data["size"]}{file}')

    if not dry_run and not leave_empty_dirs:
        remove_empty_directories(path, verbose)


@cli.command()
@click.option(
    "--path",
    "-p",
    type=click.STRING,
    default=str(),
    help="path to look in for duplicates, if this option is present, paths is ignored",
)
@click.option(
    "--paths_file_stdin",
    "-in",
    type=click.File("r"),
    default=sys.stdin,
    help="either a file containing a list of paths to evaluate or stdin",
)
@click.option(
    "--hash",
    "-h",
    type=click.STRING,
    default="sha256",
    help=f"the hashalgorithm to use, default = sha256, allowed alogorithsm: {get_available_hashes()}",
)
@click.option("--depth_min", "-d", is_flag=True, help="keep the file with the lowest pathway depth")
@click.option("--depth_max", "-D", is_flag=True, help="keep the file with the highest pathway depth")
@click.option("--name_min", "-n", is_flag=True, help="keep the file with the shortest name")
@click.option("--name_max", "-N", is_flag=True, help="keep the file with the longest name")
@click.option("--created_min", "-c", is_flag=True, help="keep the file with the oldest creation date")
@click.option("--created_max", "-C", is_flag=True, help="keep the file with the newest creation date")
@click.option("--modified_min", "-m", is_flag=True, help="keep the file with the oldest modified date")
@click.option("--modified_max", "-M", is_flag=True, help="keep the file with the newest modified, date")
@click.option("--accessed_min", "-a", is_flag=True, help="keep the file with the oldest accessed, date")
@click.option("--accessed_max", "-A", is_flag=True, help="keep the file with the newest accessed, date")
@click.option(
    "--number_of_cpus", "-ncpu", type=click.INT, default=get_max_workers(), help="maximum number of cpu cores to use"
)
@click.option("--chunksize", "-ch", type=click.INT, default=2, help="chunksize to give to workers, minimum of 2")
@log_func_start_finish_flags
@log_func_with_args
def scan(
    path: str,
    paths_file_stdin: click.File,
    hash: str,
    depth_min: bool,
    depth_max: bool,
    name_min: bool,
    name_max: bool,
    created_min: bool,
    created_max: bool,
    modified_min: bool,
    modified_max: bool,
    accessed_min: bool,
    accessed_max: bool,
    number_of_cpus: int = get_max_workers(),
    chunksize: int = 2,
) -> None:
    """
    Scan recursively computes a hash of each file and puts the hash into
    a dictionary.  The keys are the hashes of the files, and the values
    are the file paths and metadata.  If an entry has more than 1 file
    associated, they are duplicates.  The original is determined by the
    flags or options (ex: -d).  The duplicates are added to a file called
    duple.delete.
    """
    start = perf_counter()

    # input validation
    hashes = get_available_hashes()
    if hash not in hashes:
        click.secho(f"Hash must be one of the following: {hashes}")
        return

    if number_of_cpus == 0 or number_of_cpus > os.cpu_count():
        number_of_cpus = get_max_workers()
        click.echo(f"Invalid number_of_cpus, using {number_of_cpus}")

    if chunksize < 2:
        chunksize = 2
        click.secho(f"chunksize too low, setting to default of {chunksize}")

    flags = [
        depth_min,
        depth_max,
        name_min,
        name_max,
        accessed_min,
        accessed_max,
        created_min,
        created_max,
        modified_min,
        modified_max,
    ]

    c = Counter(flags)
    if c[True] == 0:
        click.secho("Must select at least one flag to determine handling of duplicates, ex: -d")

    if depth_min and depth_max:
        click.secho(f"{depth_min=} and {depth_max=}, Only one depth flag (-d or -D) can be set")
        return

    if name_min and name_max:
        click.secho(f"{name_min=} and {name_max=}, Only one name flag (-n or -N) can be set")
        return

    if created_min and created_max:
        click.secho(f"{created_min=} and {created_max=}, Only one created flag (-c or -C) can be set")
        return

    if modified_min and modified_max:
        click.secho(f"{modified_min=} and {modified_max=}, Only one modified flag (-m or -M) can be set")
        return

    logger.debug("Finished input validation")

    summary_statistics = dict()
    summary_statistics["Total Files"] = ""
    summary_statistics["Ignored Files"] = ""
    summary_statistics["Duplicate Files"] = ""
    summary_statistics["Duplicate Groups"] = ""
    summary_statistics["Total Size (duplicates)"] = ""
    summary_statistics["Total Size (all files)"] = ""
    summary_statistics["Hash Algorithm"] = hash
    summary_statistics["File System Traversal Time (seconds)"] = ""
    summary_statistics["Pre-Processing Files Time (seconds)"] = ""
    summary_statistics["Hashing Time (seconds)"] = ""
    summary_statistics["Total Time (seconds)"] = ""
    summary_statistics["Duple Version"] = __version__
    summary_statistics["Results Written To"] = ""
    summary_statistics["Total Time (seconds)"] = start

    # create filtering options list
    options = generate_options_list(flags)

    # file traversal
    summary_statistics["File System Traversal Time (seconds)"] = perf_counter()
    if path == "":
        path = os.getcwd()
        with paths_file_stdin:
            filelist = paths_file_stdin.read()
            filelist = filelist.split("\n")[:-1]
        logger.debug(f"Finished reading file list from stdin, {len(filelist)} files")
    else:
        filelist = list()
        for r, fds, fs in tqdm(os.walk(path, followlinks=False), desc="traversing file tree"):
            for f in fs:
                filelist.append(Path(r).joinpath(f).absolute())

        logger.debug(f"Finished reading files from os.walk, {len(filelist)} files")

    summary_statistics["File System Traversal Time (seconds)"] = (
        perf_counter() - summary_statistics["File System Traversal Time (seconds)"]
    )

    # create files
    summary_statistics["Pre-Processing Files Time (seconds)"] = perf_counter()
    files = Files(max_workers=number_of_cpus, chunksize=chunksize, hashalgo=hash)
    files.read_paths(filelist)
    summary_statistics["Pre-Processing Files Time (seconds)"] = (
        perf_counter() - summary_statistics["Pre-Processing Files Time (seconds)"]
    )

    # analyze duplicates
    summary_statistics["Hashing Time (seconds)"] = perf_counter()
    files.perform_analysis()
    logger.debug("Finished hashing")
    files.determine_originals(options)
    logger.debug("Finished selecting originals")
    summary_statistics["Hashing Time (seconds)"] = perf_counter() - summary_statistics["Hashing Time (seconds)"]

    summary_statistics["Total Files"] = files.file_count
    summary_statistics["Ignored Files"] = files.ignored_files
    summary_statistics["Duplicate Files"] = files.duplicate_file_count
    summary_statistics["Duplicate Groups"] = files.duplicate_group_count
    summary_statistics["Total Size (duplicates)"] = naturalsize(files.duplicate_file_size)
    summary_statistics["Total Size (all files)"] = naturalsize(files.total_file_size)
    summary_statistics["Total Time (seconds)"] = perf_counter() - summary_statistics["Total Time (seconds)"]
    summary_statistics["Results Written To"] = str(Path(path).joinpath("duple.delete").absolute())

    # create output
    create_output(
        options, path, files.create_duplicate_output(), files.create_ignored_files_output(), summary_statistics
    )
    logger.debug(f'Finished creating output, total "duple scan" time of {round(perf_counter() - start, 5): .5f}')


@log_func_with_args
@cli.command()
@log_func_start_finish_flags
def version():
    """display the current version of duple"""
    click.secho(f"duple version: {__version__}")



import os
import sysconfig
from pathlib import Path
import tomllib

PROJECT_ROOT = Path(__file__).parent.parent
PROJECT_TOML = PROJECT_ROOT.joinpath("pyproject.toml")
USER_SCRIPTS_PATH = sysconfig.get_path("scripts", f"{os.name}_user")
VERSION_PATH = Path(__file__).parent.joinpath("__version__.py")
APP_NAME_PATH = Path(__file__).parent.joinpath("__app_name__.py")
LOGGING_CONFIGURATION_PATH = Path(__file__).parent.joinpath("logging_configuration.json")
LOGS_PATH = Path(__file__).parent.joinpath("logs")

if Path(PROJECT_TOML).exists():
    with PROJECT_TOML.open("rb") as toml_file:
        PYPROJECT = tomllib.load(toml_file)



from duple.__app_name__ import APP_NAME
from duple.info import LOGGING_CONFIGURATION_PATH, LOGS_PATH


import atexit
from pathlib import Path

import logging

import json
import logging.config
import logging.handlers

logger = logging.getLogger(APP_NAME)


def touch(path: Path):
    if isinstance(path, str):
        path = Path(path)

    with open(path, "w"):
        return


def setup_logging():
    with open(LOGGING_CONFIGURATION_PATH) as f_in:
        config = json.load(f_in)

    log_path_jsonl = str(Path(LOGS_PATH).joinpath("log.jsonl").absolute())
    log_path_debug = str(Path(LOGS_PATH).joinpath("log.log").absolute())

    if not Path(log_path_jsonl).exists():
        touch(log_path_jsonl)
    if not Path(log_path_debug).exists():
        touch(log_path_debug)

    config["handlers"]["file_json"]["filename"] = str(Path(LOGS_PATH).joinpath("log.jsonl").absolute())
    config["handlers"]["file_debug"]["filename"] = str(Path(LOGS_PATH).joinpath("log.log").absolute())

    logging.config.dictConfig(config)

    queue_handler = logging.getHandlerByName("queue_handler")
    if queue_handler is not None:
        queue_handler.listener.start()
        atexit.register(queue_handler.listener.stop)


def main():
    setup_logging()
    logging.basicConfig(level="INFO")
    logger.debug("debug message", extra={"x": "hello"})
    logger.info("info message")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")
    try:
        1 / 0
    except ZeroDivisionError:
        logger.exception("exception message")


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from pathlib import Path
import hashlib
from duple.status import Status
from duple.dispocode import DispoCode
from duple.app_logging import logger


@dataclass
class File:
    path: Path
    status: Status = field(init=False, default=Status.NOT_ANALYZED)
    name: str = field(init=False, default="")
    namelength: int = field(init=False, default=0)
    twins: list = field(init=False, default=0)
    size: int = field(init=False, default=0)
    atime: float = field(init=False, default=0)
    ctime: float = field(init=False, default=0)
    mtime: float = field(init=False, default=0)
    hash: str = field(init=False, default="")
    depth: str = field(init=False, default=0)
    message: str = field(init=False, default="")
    dispocode: DispoCode = field(init=False, default=DispoCode.NOT_SET)

    def get_available_option_attributes():
        result = ["depth", "namelength", "atime", "ctime", "mtime"]
        return result

    def calc_hash(self, hashalgo: str = ""):
        with open(self.path, "rb") as f:
            digest = hashlib.file_digest(f, hashalgo)
        self.hash = digest.hexdigest()
        return self

    def __post_init__(self):
        if isinstance(self.path, str):
            self.path = Path(self.path)

        self.depth = len(self.path.parents)
        self.name = self.path.name
        self.namelength = len(self.name)

        if self.path.is_file():
            self.size = self.path.stat().st_size
            self.atime = self.path.stat().st_atime
            self.ctime = self.path.stat().st_ctime
            self.mtime = self.path.stat().st_mtime
        else:
            self.status = Status.IGNORED
            self.message = "File Path Does Not Exist or Unrecognized Characters in File Path/Name"
            logger.info(f"File Ignored, pathlib.Path.is_file() == False, {str(self.path.absolute())}")
